package es.uca.xmleye.formats;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Map;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import javax.xml.xpath.*;

import es.uca.xmleye.util.CommandBuilder;
import es.uca.xmleye.xpath.HashNamespaceContext;

/**
 * FormatDescriptor de documento para el que se tiene soporte.
 * 
 * En las órdenes, se utiliza <code>%s</code> para indicar el fichero fuente.
 * Se sustituirá con la ruta completa correcta cuando sea necesario.
 * 
 * Estos documentos son ficheros XML sencillos, del formato:
 * 
 * <pre>
 *  &lt;![CDATA[
 *  &lt;format&gt;
 *  &lt;!-- Nombre por defecto --&gt;
 *  &lt;name&gt;Mi Formato&lt;/name&gt;
 *  &lt;!-- Nombre en inglés (Estados Unidos) --&gt;
 *  &lt;name language=&quot;en_US&quot;&gt;My Formato&lt;/name&gt;
 *  &lt;edit_cmd&gt;mi-editor-favorito %s&lt;/edit_cmd&gt;
 *  &lt;import_cmd&gt;mi-preproc-favorito %s&lt;/import_cmd&gt;
 *  &lt;extensions&gt;
 *  &lt;extension&gt;abc&lt;/extension&gt;
 *  &lt;extension&gt;def&lt;/extension&gt;
 *  &lt;/extensions&gt;
 *  &lt;/formato&gt;
 *  ]]&gt;
 * </pre>
 * 
 * Esta clase provee los métodos necesarios para acceder a dicha información de
 * manera directa, sin que sea necesario manipular el código XML.
 * 
 * Para más detalles, consultar el XML Schema dentro del directorio
 * <code>src/formats</code>.
 * 
 * La orden de importación es opcional, cubriendo así los casos en que
 * trabajemos directamente con ficheros XML.
 * 
 * @author Antonio García Domínguez
 * @version 1.0
 */
public class FormatDescriptor {
	
	/**
	 * Clave a usar para sustituir en las órdenes por la ruta completa a la
	 * fuente del documento XML.
	 */
	public final static String CLAVE_FUENTE = "%s";
	
	final protected static String FORMATO_NS_PREFIX = "d";
	final protected static String FORMATO_NS_URI = "http://xmleye.uca.es/xmleye/accepted-doc";

	final protected static String TAG_FORMATO = "format";
	final protected static String TAG_NOMBRE = "name";
	final protected static String TAG_ORDEN_EDICION = "edit_cmd";
	final protected static String TAG_ORDEN_IMPORTACION = "import_cmd";
	final protected static String TAG_EXTENSIONES = "extensions";
	final protected static String TAG_EXTENSION = "extension";

	final protected static String ATTR_IDIOMA = "language";

	// Usado para internacionalizar los mensajes de error
	protected static final ResourceBundle errRes;

	static {
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
	}

	private File[] fuentes;
	private String patronOrdenEdicion;
	private String patronOrdenImportacion;
	private String[] extensiones;

	// Los nombres son localizables por país e idioma
	private Map<String, String> mapNombres;
	// Se notifica al padre cuando cambia el nombre
	private FormatDescriptorRepository padre;

	/**
	 * Construye el formato cargándolo desde un fichero.
	 * 
	 * Método de conveniencia para <code>FormatDescriptor(File[])</code>.
	 * 
	 * @param padre
	 *            Repositorio en que se halla contenido este formato.
	 * @param fuente
	 *            Fichero desde el que cargar los datos.
	 * @return FormatDescriptor con los datos cargados.
	 * @exception FormatDescriptorException
	 *                No se pudieron cargar los datos.
	 */
	FormatDescriptor(FormatDescriptorRepository padre, File fuente)
			throws FormatDescriptorException {
		cargar(padre, new File[] { fuente });
	}

	/**
	 * Construye el formato intentando cargarlo de un conjunto de ficheros.
	 * 
	 * @param padre
	 *            Repositorio en que se halla contenido este formato.
	 * @param fuentes
	 *            Lista de posibles fuentes en orden descendente de prioridad.
	 * @return FormatDescriptor con los datos cargados.
	 * @exception FormatDescriptorException
	 *                No se pudieron cargar los datos.
	 */
	FormatDescriptor(FormatDescriptorRepository padre, File[] fuentes)
			throws FormatDescriptorException {
		cargar(padre, fuentes);
	}

	/** METODOS PUBLICOS * */

	/** * ACCESORES ** */
	public ArrayList<String> getOrdenImportacion(File fich) {
		return instanciarPatronOrden(fich, getPatronOrdenImportacion());
	}

	public ArrayList<String> getOrdenEdicion(File fich) {
		return instanciarPatronOrden(fich, getPatronOrdenEdicion());
	}
	
	public String getPatronOrdenEdicion() {
		return patronOrdenEdicion;
	}

	public String getPatronOrdenImportacion() {
		return patronOrdenImportacion;
	}

	public String getNombre() {
		final String idioma = Locale.getDefault().getLanguage();
		final String pais = Locale.getDefault().getCountry();

		// Claves a ir probando
		final String[] claves = new String[] { Locale.getDefault().toString(),
				idioma + "_" + pais, idioma };

		for (String k : claves) {
			if (mapNombres.containsKey(k))
				return mapNombres.get(k);
		}
		return mapNombres.get("");
	}

	public String[] getExtensiones() {
		return extensiones;
	}

	/** * MUTADORES ** */

	public void setPatronOrdenEdicion(String patronOrden) throws FormatDescriptorException {
		if (patronOrden == null || patronOrden.trim().equals("")) {
			throw new FormatDescriptorException(errRes
					.getString("err_invalid_editcmd"));
		}
		this.patronOrdenEdicion = patronOrden;
	}

	public void setPatronOrdenImportacion(String patronOrden)
			throws FormatDescriptorException {
		if (patronOrden == null || (!patronOrden.equals("") && patronOrden.trim().equals(""))) {
			throw new FormatDescriptorException(errRes
					.getString("err_invalid_importcmd"));
		}
		this.patronOrdenImportacion = patronOrden;
	}

	public void setNombre(String nombre) throws FormatDescriptorException {
		if (nombre == null || nombre.trim().equals("")) {
			throw new FormatDescriptorException(errRes
					.getString("err_invalid_name"));
		}
		// Usamos siempre la coincidencia más exacta
		mapNombres.put(Locale.getDefault().toString(), nombre);
		padre.nombreCambiado(this);
	}

	public void setExtensiones(String[] exts) throws FormatDescriptorException {
		if (exts == null || exts.length == 0) {
			throw new FormatDescriptorException(errRes
					.getString("err_invalid_exts"));
		}
		for (String s : exts) {
			// No extension should contain spaces
			if (!s.matches("\\S+")) {
				throw new FormatDescriptorException(errRes
						.getString("err_invalid_exts"));
			}
		}
		this.extensiones = exts;
	}

	/** * CONVERSIÓN A CADENA ** */
	@Override
	public String toString() {
		return getNombre();
	}

	/**
	 * Retira una personalización del formato, sin llegar a borrarlo.
	 * 
	 * El proceso consiste simplemente en borrar el primer fichero de formato en
	 * prioridad, tras comprobar que existe un segundo, y volver a cargar los
	 * datos.
	 * 
	 * @exception FormatDescriptorException
	 *                No se ha podido eliminar la versión personalizada.
	 */
	public void restaurar() throws FormatDescriptorException {
		try {
			int posPrimero;
			int posSegundo;

			// Buscamos la versión segunda en prioridad
			for (posPrimero = 0; posPrimero < fuentes.length
					&& !fuentes[posPrimero].canRead(); ++posPrimero)
				;

			for (posSegundo = posPrimero + 1; posSegundo < fuentes.length
					&& !fuentes[posSegundo].canRead(); ++posSegundo)
				;

			// Existe: borramos la más importante
			if (posSegundo < fuentes.length) {
				fuentes[posPrimero].delete();
			}

			// Cargamos de nuevo los datos
			cargar(padre, fuentes);
		} catch (SecurityException ex) {
			throw new FormatDescriptorException(ex);
		}
	}

	/** METODOS DE PAQUETE * */

	/**
	 * Carga desde la posible fuente útil con más prioridad.
	 * 
	 * @param padre
	 *            Repositorio padre al que notificar cambios en el nombre.
	 * @param fuentes
	 *            Lista de posibles ficheros fuente, ordenados en prioridad
	 *            descendente.
	 * @exception FormatDescriptorException
	 *                Hubo algún problema al cargar.
	 */
	void cargar(FormatDescriptorRepository padre, File[] fuentes)
			throws FormatDescriptorException {
		this.padre = padre;
		this.fuentes = fuentes;

		for (File f : fuentes) {
			if (f.canRead()) {
				cargar(f);
				return;
			}
		}

		throw new FormatDescriptorException(errRes
				.getString("err_cannot_open_format"));
	}

	/**
	 * Carga los datos desde el fichero <code>fuente</code>.
	 * 
	 * El fichero <code>fuente</code> pasará a usarse en el próximo guardado.
	 * 
	 * @param fuente
	 *            Fichero desde el que cargar los datos.
	 * @exception FormatDescriptorException
	 *                Hubo algún problema al cargar.
	 */
	void cargar(File fuente) throws FormatDescriptorException {
		try {
			// Usaremos un parser que valide la entrada con un XML Schema
			DocumentBuilderFactory docBF = DocumentBuilderFactory.newInstance();
			docBF.setNamespaceAware(true);
			docBF.setValidating(false);
			try {
				docBF.setSchema(padre.getSchema());
			} catch (java.io.FileNotFoundException ex) {
				System.err.println("(!!)\t" + ex.getMessage());
			}

			// Analizamos la entrada
			DocumentBuilder docB = docBF.newDocumentBuilder();
			Document doc = docB.parse(fuente);

			// Emplearemos XPath para extraer la información que
			// necesitamos (hay que ajustar el nombre de espacios)
			XPath xpath = XPathFactory.newInstance().newXPath();
			HashNamespaceContext nsContext = new HashNamespaceContext();
			nsContext.set(FORMATO_NS_PREFIX, FORMATO_NS_URI);
			xpath.setNamespaceContext(nsContext);

			this.mapNombres = new HashMap<String, String>();
			NodeList listaNombres = (NodeList) xpath.evaluate(String.format(
					"/p:%s/p:%s".replaceAll("p(?=:)", FORMATO_NS_PREFIX),
					TAG_FORMATO, TAG_NOMBRE), doc, XPathConstants.NODESET);
			for (int i = 0; i < listaNombres.getLength(); ++i) {
				Element elem = (Element) listaNombres.item(i);
				String idioma = elem.getAttribute(ATTR_IDIOMA);
				mapNombres.put(idioma != null ? idioma : "", elem
						.getFirstChild().getNodeValue());
			}

			this.patronOrdenEdicion = xpath
					.evaluate(String.format("/p:%s/p:%s/text()".replaceAll(
							"p(?=:)", FORMATO_NS_PREFIX), TAG_FORMATO,
							TAG_ORDEN_EDICION), doc);
			this.patronOrdenImportacion = xpath.evaluate(String
					.format("/p:%s/p:%s/text()".replaceAll("p(?=:)",
							FORMATO_NS_PREFIX), TAG_FORMATO,
							TAG_ORDEN_IMPORTACION), doc);
			NodeList listaExts = (NodeList) xpath.evaluate(String.format(
					"/p:%s/p:%s/p:%s/text()".replaceAll("p(?=:)",
							FORMATO_NS_PREFIX), TAG_FORMATO, TAG_EXTENSIONES,
					TAG_EXTENSION), doc, XPathConstants.NODESET);

			this.extensiones = new String[listaExts.getLength()];
			for (int i = 0; i < listaExts.getLength(); ++i) {
				this.extensiones[i] = listaExts.item(i).getNodeValue();
			}

		} catch (IOException ex) {
			throw new FormatDescriptorException(ex);
		} catch (SAXException ex) {
			throw new FormatDescriptorException(ex);
		} catch (ParserConfigurationException ex) {
			throw new FormatDescriptorException(ex);
		} catch (XPathExpressionException ex) {
			throw new FormatDescriptorException(ex);
		}
	}

	/**
	 * Guarda en la fuente original con más prioridad.
	 * 
	 * @exception FormatDescriptorException
	 *                Hubo algún problema al guardar.
	 */
	void guardar() throws FormatDescriptorException {
		guardar(fuentes);
	}

	/**
	 * Guarda en el fichero con más prioridad escribible de
	 * <code>destinos</code>.
	 * 
	 * @param destinos
	 *            Lista de posibles destinos ordenada descendentemente por
	 *            prioridad.
	 * @exception FormatDescriptorException
	 *                Hubo algún problema al guardar.
	 */
	void guardar(File[] destinos) throws FormatDescriptorException {
		try {
			for (File f : destinos) {
				if (f.canWrite() || f.createNewFile()) {
					guardar(f, destinos[destinos.length - 1].getAbsoluteFile()
							.getParent());
					return;
				}
			}
		} catch (IOException ex) {
			throw new FormatDescriptorException(ex);
		}

		throw new FormatDescriptorException(errRes
				.getString("err_cannot_save_format"));
	}

	/**
	 * Guarda los cambios en el fichero <code>destino</code>.
	 * 
	 * @param destino
	 *            Fichero al que guardar los datos.
	 * @param dirDTD
	 *            Directorio en el que se halla la DTD.
	 * @exception FormatDescriptorException
	 *                Hubo algún problema al guardar.
	 */
	void guardar(File destino, final String dirDTD)
			throws FormatDescriptorException {
		try {
			PrintWriter ow = new PrintWriter(new OutputStreamWriter(
					new BufferedOutputStream(new FileOutputStream(destino)),
					"UTF-8"));

			ow.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
			ow.println(String.format("<%s xmlns=\"%s\">", TAG_FORMATO,
					FORMATO_NS_URI));

			for (Map.Entry<String, String> elem : mapNombres.entrySet()) {
				if ("".equals(elem.getKey())) {
					ow.println(String.format("  <%s>%s</%s>", TAG_NOMBRE, elem
							.getValue(), TAG_NOMBRE));
				} else {
					ow.println(String.format("  <%s language=\"%s\">%s</%s>",
							TAG_NOMBRE, elem.getKey(), elem.getValue(),
							TAG_NOMBRE));
				}
			}

			ow.println(String.format("  <%s>%s</%s>", TAG_ORDEN_EDICION,
					getPatronOrdenEdicion(), TAG_ORDEN_EDICION));
			ow.println(String.format("  <%s>%s</%s>", TAG_ORDEN_IMPORTACION,
					getPatronOrdenImportacion(), TAG_ORDEN_IMPORTACION));
			ow.println(String.format("  <%s>", TAG_EXTENSIONES));
			for (String ext : getExtensiones()) {
				ow.println(String.format("    <%s>%s</%s>", TAG_EXTENSION, ext,
						TAG_EXTENSION));
			}
			ow.println(String.format("  </%s>", TAG_EXTENSIONES));
			ow.println(String.format("</%s>", TAG_FORMATO));

			ow.close();
		} catch (FileNotFoundException ex) {
			throw new FormatDescriptorException(ex);
		} catch (UnsupportedEncodingException ex) {
			throw new FormatDescriptorException(ex);
		}
	}

	/* METODOS PRIVADOS */
	
	private ArrayList<String> instanciarPatronOrden(File fich, String orden) {
		return CommandBuilder.prepararOrden(
				orden,
				new String[] { CLAVE_FUENTE }, 
				new String[] { fich.getAbsolutePath() });
	}

}
