package es.uca.xmleye.gui.trees;

/*
 Copyright (C) 2005-2008 Antonio Garcia Dominguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import org.w3c.dom.*;
import java.io.*;
import es.uca.xmleye.ParsingException;

/**
 * <p>
 * Extiende la lógica de DOMTree, ocultando un nodo o sus descendientes en
 * función de los datos del propio documento XML.
 * </p>
 * 
 * <p>
 * Los elementos tales que <code>HIDDEN_ATTRIBUTE ==
 ENABLED_VALUE</code> no se
 * mostrarán. Tampoco se mostrarán los descendientes de aquellos elementos tales
 * que <code>LEAF_ATTRIBUTE ==
 ENABLED_VALUE</code>.
 * </p>
 */
public class DataDrivenTreeModel extends DOMTreeModel {
	/** Atributo usado para marcar los descendientes del nodo. */
	public final static String LEAF_ATTRIBUTE = "leaf";
	/** Atributo usado para ocultar el nodo. */
	public final static String HIDDEN_ATTRIBUTE = "hidden";
	/** Valor a dar a los atributos anteriores para activarlos. */
	public final static String ENABLED_VALUE = "1";

	/**
	 * Constructor por defecto. Crea un documento vacío.
	 */
	public DataDrivenTreeModel() throws ParsingException {
		super();
	}

	/**
	 * Utiliza el árbol DOM ya existente de un documento XML.
	 * 
	 * @param d
	 *            Documento XML a usar como base.
	 */
	public DataDrivenTreeModel(Document d) {
		super(d);
	}

	/**
	 * Toma un flujo de entrada con un documento XML, genera su árbol DOM y lo
	 * usa como documento.
	 * 
	 * @param is
	 *            Flujo de entrada con el documento XML.
	 * 
	 * @throws ParsingException
	 *             Excepción durante el análisis del código XML.
	 */
	public DataDrivenTreeModel(InputStream is) throws ParsingException {
		super(is);
	}

	/**
	 * Además de la lógica de DOMTreeModel, nodos hoja son también aquellos tales
	 * que <code>LEAF_ATTRIBUTE == ENABLED_VALUE</code>.
	 * 
	 * @param nodo
	 *            Nodo en consideración.
	 * @return <code>true</code> si debe ser considerado hoja,
	 *         <code>false</code> si no.
	 */
	@Override
	public boolean isLeaf(Object nodo) {
		return isLeafImpl((Node) nodo);
	}

	/**
	 * Además de la lógica de DOMTreeModel, un nodo tal que
	 * <code>LEAF_ATTRIBUTE == ENABLED_VALUE</code> tendrá siempre 0 hijos.
	 * 
	 * @param padre
	 *            Nodo en consideración.
	 * @return Número de hijos de <code>padre</code> o 0 si debe ser
	 *         considerado hoja.
	 */
	@Override
	public int getChildCount(Object padre) {
		return isLeafImpl((Node) padre) ? 0 : super.getChildCount(padre);
	}

	@Override
	public Object getChild(Object padre, int idHijo) {
		return isLeafImpl((Node) padre) ? null : super.getChild(padre, idHijo);
	}

	@Override
	public int getIndexOfChild(Object padre, Object hijo) {
		return isLeafImpl((Node) padre) ? -1 : super.getIndexOfChild(padre,
				hijo);
	}

	/** * METODOS PROTEGIDOS ************************** */

	/**
	 * Aquí retiraremos de la vista nodos tales que <code>HIDDEN_ATTRIBUTE
	    == ENABLED_VALUE</code>.
	 * 
	 * @param n
	 *            Nodo a considerar.
	 * @return <code>true</code> si debe aparecer en la vista,
	 *         <code>false</code> si no.
	 */
	@Override
	protected boolean filterNode(Node n) {

		/*
		 * NOTA: la evaluación con cortocircuito permite evitar problemas a la
		 * hora de evaluar nodos que no son elementos (super.filtrarNodo()
		 * devolverá falso en dicho caso)
		 */

		return super.filterNode(n)
				&& !((Element) n).getAttribute(HIDDEN_ATTRIBUTE).equals(
						ENABLED_VALUE);
	}

	/**
	 * Realiza un filtrado, ocultando los hijos de un nodo con
	 * <code>LEAF_ATTRIBUTE == ENABLED_VALUE</code>, para interés de
	 * visualización.
	 * 
	 * @param elem
	 *            Nodo a filtrar.
	 * @return <code>true</code> si se va a considerar como hoja,
	 *         <code>false</code> si no.
	 */
	private boolean isLeafImpl(Node elem) {

		/*
		 * NOTA: la evaluación con cortocircuito permite evitar problemas a la
		 * hora de evaluar nodos que no son elementos (super.esHoja() devolverá
		 * verdadero en dicho caso)
		 */

		return super.isLeaf(elem)
				|| ((Element) elem).getAttribute(LEAF_ATTRIBUTE).equals(
						ENABLED_VALUE);
	}

}
