package es.uca.xmleye.xslt;

/*
  Copyright (C) 2008 Antonio García Domínguez
  $Id$

  This program is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public
  License along with this program; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA

*/

import java.io.*;
import java.util.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.Configuration;
import net.sf.saxon.TransformerFactoryImpl;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.trans.XPathException;

/**
 *
 * Stylesheet de transformación XSLT con URIResolver personalizado, salidas en
 * forma de árbol DOM o fuente XML e inicialización tardía.
 *
 */
public class Stylesheet {
  // Usado para internacionalizar los mensajes de error
  protected static final ResourceBundle errRes;

  static {
    errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
                                      Locale.getDefault());

    // Use Saxon 9.1 (XSLT 2.0) as the XSLT engine
    System.setProperty("javax.xml.transform.TransformerFactory",
                       "net.sf.saxon.TransformerFactoryImpl");
  }

  /// Ruta absoluta al documento XSLT
  private String rutaXSLT;
  /// Resolvedor de URI
  private URIResolver resolvedorURI;

  /// Translet almacenado (inicialización tardía)
  private Templates translet;

  /* --- METODOS PUBLICOS / CONSTRUCTORES --- */

  /**
   * Crea una nueva hoja, usando el resolvedor por defecto.
   */
  Stylesheet(String pRutaXSLT) {
    this.rutaXSLT = pRutaXSLT;
    this.resolvedorURI = null;
  }

  /**
   * Crea una nueva hoja, con un resolvedor personalizado.
   */
  Stylesheet(String pRutaXSLT, URIResolver pResolvedorURI) {
    this.rutaXSLT = pRutaXSLT;
    this.resolvedorURI = pResolvedorURI;
  }

  /**
   * Transforma el documento en <code>fuente</code> al destino
   * <code>resultado</code>.
   */
  public void transformar(Source fuente, Result resultado)
      throws TransformationException {
    try {
      final Transformer transformador = getTranslet().newTransformer();

      // Para subclases
      configurarTransformador(transformador);

      transformador.transform(fuente, resultado);
    } catch (TransformerConfigurationException transConfEx) {
      throw new TransformationException(
          errRes.getString("err_conf_transformer")
          + "\n"
          + transConfEx.getLocalizedMessage(),
          transConfEx);
    } catch (TransformerException transEx) {
      throw new TransformationException(
          errRes
          .getString("err_proc_transformer")
          + "\n"
          + transEx.getLocalizedMessage(),
          transEx);
    }

  }

  /**
   * Devuelve la ruta absoluta al documento XSLT.
   */
  public String getRutaXSLT() {
    return this.rutaXSLT;
  }

  /* --- METODOS PROTEGIDOS --- */

  /**
   * Configura el transformador previamente a su uso.
   */
  protected void configurarTransformador(Transformer t)
      throws TransformerConfigurationException {
  }

  /* --- METODOS PRIVADOS --- */

  private Templates crearTranslet() throws TransformationException {
    try {
      // Cargamos la fuente del documento XSLT
      final InputStream isXSLT
          = new BufferedInputStream(new FileInputStream(rutaXSLT));

      // Preparamos la fábrica
      final TransformerFactory fabrica = TransformerFactory.newInstance();
      if (resolvedorURI != null)
        fabrica.setURIResolver(resolvedorURI);
      fabrica.setAttribute(net.sf.saxon.FeatureKeys.LINE_NUMBERING, true);

      // Compilamos el transformador a un translet
      Configuration saxonConfiguration
          = ((TransformerFactoryImpl) fabrica).getConfiguration();
      NodeInfo xsltDoc
          = saxonConfiguration.buildDocument(
              new StreamSource(isXSLT));
      return fabrica.newTemplates(xsltDoc);
    }
    catch (XPathException xpathEx) {
      throw new TransformationException(
          errRes.getString("err_eval_xpathexpr_1")
          + "\n"
          + xpathEx.getLocalizedMessage(),
          xpathEx);
    }
    catch (TransformerConfigurationException transConfEx) {
      throw new TransformationException(
          errRes.getString("err_conf_transformer")
          + "\n"
          + transConfEx.getLocalizedMessage(),
          transConfEx);
    } catch (FileNotFoundException fEx) {
      throw new TransformationException(
          String.format(errRes
                        .getString("err_xslnotfound"), rutaXSLT), fEx);
    }
  }

  synchronized Templates getTranslet() throws TransformationException {
    if (translet == null) {
      translet = crearTranslet();
    }
    return translet;
  }
}
